// ------------------------------- //
// -------- Start of File -------- //
// ------------------------------- //
// ----------------------------------------------------------- //
// C++ Header File Name: vbdfile.hpp 
// Compiler Used: MSVC40, DJGPP 2.7.2.1, GCC 2.7.2.1, HP CPP 10.24
// Produced By: Doug Gaer    
// File Creation Date: 02/04/97 
// Date Last Modified: 12/03/97
// Copyright (c) 1997 Douglas M. Gaer
// --------------------------------------------------------- // 
// ---------- Include File Description and Details  ---------- // 
// ----------------------------------------------------------- // 
/*
(V)ariable (B)lock (D)atabase Version 1024, Revision 0

The VBD C++ classes are copyright (c) 1997, by Douglas M. Gaer.
All those who put this code or its derivatives in a commercial
product MUST mention this copyright in their documentation for
users of the products in which this code or its derivative
classes are used. Otherwise, you have the freedom to redistribute
verbatim copies of this source code, adapt it to your specific
needs, or improve the code and release your improvements to the
public provided that the modified files carry prominent notices
stating that you changed the files and the date of any change.

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND.
THE ENTIRE RISK OF THE QUALITY AND PERFORMANCE OF THIS SOFTWARE
IS WITH YOU. SHOULD ANY ELEMENT OF THIS SOFTWARE PROVE DEFECTIVE,
YOU WILL ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR, OR
CORRECTION.

(V)ariable (B)lock (D)atabase (V.B.D.) technology is designed to
store variable length binary data of any type. All VBD files are
created and managed by the VBD file manager class (VBDFile). The
VBDFile class is both application and platform independent. It
can be used to manage persistent objects (objects that save their
state between program invocations) or any other data type.    

Applications that write data to a VBD file must first allocate
space in the file for length of data and then store it at the
file address passed back by the VBD file manager's allocation
function. The VBD file manger stores a (V)ariable (B)lock header
with each variable block that is allocated by the application. VB
headers are used by the VBD file manager to keep track of the data
stored in the VBD file. Applications can also use VB header info
to obtain file locations and status information of its data stored
in the VBD file. VBDFile objects are reference counted to detect
synchronization problems when closing files, and must be created
dynamically. 

VBD files gain their platform independence by using its own
representation of 32 bit signed integers for each file address. 
The INT32 class is used for this purpose. INT32 objects represent
32 bit signed integers independently of the operating system or
hardware platform used by manipulating long integer values in
memory. This allows the VBD file manager to overcome the big and
little endian problems encountered when writing long integer
values to a common database file accessed by several different
types of machines.
*/
// ----------------------------------------------------------- //   
#ifndef __VBDFILE_HPP
#define __VBDFILE_HPP

#include <stdio.h>
#include "refcount.hpp"
#include "ehandler.hpp"
#include "dtypes.hpp"
#include "int32.hpp"

// typedefs for file operations and stream position
typedef INT32 FAU;       // (F)ile (A)ddress (U)nit, physical file address type
typedef LWORD StreamPos; // (S)tream (P)osistion

// Variable Block Database (F)ile (H)eader
struct FileHeader // VBD file header information (28 bytes total)
{ 
  FAU FreeSpace;      // Address to first block of free heap space (4 bytes)
  FAU EndOfFile;      // Address of byte after end of file (4 bytes)
  FAU HeapStart;      // Address of the start of the heap space (4 bytes)
  FAU HighestVB;      // Highest allocated (V)ariable (B)lock
  SBYTE Signature[8]; // Signature used for every VBD file (8 bytes)
  INT32 Version;      // VBD Version number (4 bytes)
};

// (V)ariable (B)lock (H)eader 
struct VBHeader // Marks each variable data block (16 bytes total)
{
  INT32 CkWord;       // (C)heck (W)ord for file integrity checks (4 bytes)
  INT32 Length;       // Length of this block including VB header (4 bytes) 
  INT32 Status;       // Status of dynamic data stored in VB (4 byte)
  FAU NextDeletedVB;  // Pointer to next deleted VB (4 bytes) 
};

// Constants for dynamic data attributes used by the VB Header Status member
const SBYTE NormalVB  = 'N'; // Normal Read/Write attribute 
const SBYTE DeletedVB = 'D'; // Deleted Variable Block 
const SBYTE RemovedVB = 'R'; // Removed Variable Block

// Constants for file operations and stream position 
const FAU StartOfFile  = 0;      // First byte in the file
const FAU CurrAddress  = -1;     // Indicates current location
const INT32 CheckWord  = 0xfefe; // 16 bit Check Word

class VBDFile : public CountedObject
{
public:
  enum IO_Operation { READ, WRITE, SEEK };
  enum AccessMode { READWRITE, READONLY };

public:
  VBDFile();
  virtual ~VBDFile();

public:
  virtual int Create(const char *FName, FAU StaticSize = 0);
  virtual int Open(const char *FName, AccessMode Mode = READWRITE);
  virtual void Close(int flush = 1);
  virtual void Flush();
  FAU Alloc(unsigned Bytes);
  FAU ReAlloc(FAU Address, unsigned Bytes); // Reallocate space for the object
  int Delete(FAU Address); 
  int Remove(FAU Address);
  int UnDelete(FAU Address);
  void Read(void *buf, unsigned Bytes, FAU Address = CurrAddress);
  void Write(const void *buf, unsigned Bytes, FAU Address = CurrAddress);
  void Seek(FAU Offset, int SeekMode = SEEK_SET);
  int IsOpen() const { return (Status & 0x02); }
  int ReadOnly() const { return (Status & 0x04) == 0; }
  int ReadyForWriting() const { return (Status & 0x07) == 0x07; }
  void ClearErr() { Error->ClearException(); Status |= 0x01; }
  int IsOK() const { return (Status & 0x03) == 0x03; }
  unsigned ObjectLength(FAU Address);
  unsigned VBLength(FAU Address);
  static int Exists(const char *FName);
  static LWORD FileSize(const char *FName);
  StreamPos FilePosition();
  const char *GetSignature() const;
  INT32 GetVersion() const { return Header.Version; }
  FAU GetFreeSpace() const { return Header.FreeSpace; }
  FAU GetEOF() const { return Header.EndOfFile; }
  FAU GetHeapStart() const { return Header.HeapStart; }
  FAU GetHighestVB() const { return Header.HighestVB; }
  size_t FileHeaderSize() const { return sizeof(FileHeader); }
  size_t VBHeaderSize() const { return sizeof(VBHeader); }
  unsigned VBDeleted(); // Return total number of removed and deleted VBs
  unsigned VBTotal();   // Return total number of Variable Blocks
  const char *VBDFileName() const { return (const char *)FileName; }
  LWORD StaticArea() const { return Header.HeapStart - sizeof(FileHeader); } 
  SBYTE GetFileStatus() const { return Status; }
  FAU FindFirstVB(FAU Offset = 0);     // Finds first valid Variable Block
  FAU FindNextVB(FAU Offset = 0);      // Finds next after first 
  FAU FindFirstObject(FAU Offset = 0); // Finds first object
  FAU FindNextObject(FAU Offset = 0);  // Finds next after first 
  
protected:
  FAU Reclaim(unsigned Bytes); 
  void InitHdr();
  void ReadHdr() { Read(&Header, sizeof(FileHeader), StartOfFile); }
  void WriteHdr() { Write(&Header, sizeof(FileHeader), StartOfFile); }
  void ReadVBHdr(VBHeader &hdr, FAU Address);
  void WriteVBHdr(const VBHeader &hdr, FAU Address);

protected:
  char FileName[255];           // Name of the file
  static SBYTE VBDSignature[8]; // Signature for VBD files
  static LWORD VBDVersion;      // Version number VBD files
  FileHeader Header;            // File header as stored in memory
  FILE *fp;                     // Stream file handle
  IO_Operation LastOperation;   // Last I/O operation

  SBYTE Status; // File Status bits used in VBD Files:
  // 210
  // |||---- good = 1, bad = 0
  // |+----- open = 1, closed = 0
  // +------ read/write = 1, read/only = 0

public: // Overloaded operators
  int operator!() const { return (Status & 0x03) != 0x03; }
  operator const int () const { return (Status & 0x03) == 0x03; }
};

typedef RefCount<VBDFile> VBDFilePtr;

#endif // __VBDFILE_HPP
// ----------------------------------------------------------- // 
// ------------------------------- //
// --------- End of File --------- //
// ------------------------------- //
