// ------------------------------- //
// -------- Start of File -------- //
// ------------------------------- //
// ----------------------------------------------------------- // 
// C++ Source Code File Name: vbdfile.cpp 
// Compiler Used: MSVC40, DJGPP 2.7.2.1, GCC 2.7.2.1, HP CPP 10.24
// Produced By: Doug Gaer  
// File Creation Date: 02/04/97  
// Date Last Modified: 12/04/97
// Copyright (c) 1997 Douglas M. Gaer
// ----------------------------------------------------------- // 
// ------------- Program Description and Details ------------- // 
// ----------------------------------------------------------- // 
/*
The VBD C++ classes are copyright (c) 1997, by Douglas M. Gaer.
All those who put this code or its derivatives in a commercial
product MUST mention this copyright in their documentation for
users of the products in which this code or its derivative
classes are used. Otherwise, you have the freedom to redistribute
verbatim copies of this source code, adapt it to your specific
needs, or improve the code and release your improvements to the
public provided that the modified files carry prominent notices
stating that you changed the files and the date of any change.

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND.
THE ENTIRE RISK OF THE QUALITY AND PERFORMANCE OF THIS SOFTWARE
IS WITH YOU. SHOULD ANY ELEMENT OF THIS SOFTWARE PROVE DEFECTIVE,
YOU WILL ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR, OR
CORRECTION.

This is a (D)ebug version of the (V)ariable (B)lock (D)atabase
file manager used to debug VBD files created by the VBD file
manager.
*/
// ----------------------------------------------------------- // 
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "dvbdfile.hpp"

// Allocate all necessary static data
static char ClosedFileName[255] = "closed-file";

// Current VBD File version
LWORD VBDFile::VBDVersion = 1024;

// Last byte reserved for future revision numbers
SBYTE VBDFile::VBDSignature[8] = {'V', 'B', 'D', 'F', 'I', 'L', 'E', ' '};

VBDFile::VBDFile()
// Creates a VBD file object, with refcount of one.
{
  strcpy(FileName, ClosedFileName);
  fp = 0;
  Status = 0x01; // good, read-only, and closed
}

VBDFile::~VBDFile()
{
  Close();
}

int VBDFile::Create(const char *FName, FAU StaticSize)
// Creates and opens a new file named FName, truncating it
// if it already exists. The area at the front of the file
// of length StaticSize + sizeof(FileHeader) is reserved.
{
  // First, close the current file if open.
  Close();

  Status = 0x05; // Set read/write bit and good bit

  // Create, truncate if already exists
  fp = ::fopen(FName, "w+b");

  if (fp == 0) {
#ifdef CPP_EXCEPTIONS
    throw CFileCreationError();
#else
    Error->SignalException(EHandler::FileCreationError);
#endif
  }
  else {
    Status |= 2; // Set open bit
    strcpy(FileName, FName);
    Header.HeapStart = StaticSize + sizeof(FileHeader);
    LastOperation = READ; // So Write() works right the first time
    InitHdr();
  }

  // Returns 1 if the file was successfully created and opened.
  return IsOpen(); 
}

void VBDFile::InitHdr()
// Sets up the header area in the file. If there's
// statically allocated data beyond the header, write
// a 0 to the last byte of it, so that there will be no
// unexpected end of file errors. The rest of the static
// area stays uninitialized.
{
  Header.FreeSpace = 0;
  Header.EndOfFile = Header.HeapStart;
  Header.HighestVB = 0; // No blocks allocated in the file 
  memcpy(Header.Signature, VBDSignature, 8); // Copy signature
  Header.Version = VBDVersion;
  WriteHdr();
  if (Header.HeapStart > sizeof(FileHeader)) {
     SBYTE zero_byte = 0;
     Write(&zero_byte, 1, Header.HeapStart-1);
  }
}

int VBDFile::Open(const char *FName, AccessMode Mode)
// Opens the FName file. File must exist and be a VBDFile type 
// or error occurs. 
{
  char *mode_str;

  // First, close the current file if open.
  Close();

  if (Mode == READONLY) {
     mode_str = "rb";
     Status = 0x01; // set good bit, reset read/write bit
  }
  else {
     mode_str = "r+b";
     Status = 0x05; // Set good bit and read/write bit
  }

  fp = ::fopen(FName, mode_str);

  if (fp == 0) {
#ifdef CPP_EXCEPTIONS
    throw CFileOpenError();
#else
    Error->SignalException(EHandler::FileOpenError);
#endif
  }
  else {
    Status |= 2; // Set open bit
    strcpy(FileName, FName);
    LastOperation = WRITE; // So Read() works right the first time
    ReadHdr();
    if (memcmp(Header.Signature, VBDSignature, 8)) { // Test file type
       Close(0);
#ifdef CPP_EXCEPTIONS
       throw CWrongFileType();
#else
       Error->SignalException(EHandler::WrongFileType);
#endif
    }
  }

  return IsOpen(); // Returns 1 if file opened successfully, else 0.
}

void VBDFile::Close(int flush)
// Closes the file if not already closed, flushing the
// basic header if flush = 1. Does nothing if in the
// error state. Checks for dangling references to
// this file.
{
  if (IsOK()) {
    if (refcount > 1) // Check the object pointer's reference count
#ifdef CPP_EXCEPTIONS
      throw CDanglingPtr();
#else
    Error->SignalException(EHandler::DanglingPtr);
#endif
    
    if (flush && !ReadOnly())
      WriteHdr();
    if (::fclose(fp) != 0)
#ifdef CPP_EXCEPTIONS
      throw CFileCloseError();
#else
      Error->SignalException(EHandler::FileCloseError);
#endif
    strcpy(FileName, ClosedFileName);
  }
  Status &= 0xfd; // Reset open bit
}

void VBDFile::Flush()
// Writes the VBDFile header to the file, and then flushes
// any internal buffers the file might have.
{
  if (ReadyForWriting()) {
     WriteHdr();
     if (fflush(fp) != 0)
#ifdef CPP_EXCEPTIONS
       throw CFileWriteError();
#else
       Error->SignalException(EHandler::FileWriteError);
#endif
     Seek(0, SEEK_CUR); // So that the LastOperation logic works right
  }
}

void VBDFile::Seek(FAU Offset, int SeekMode)
// Moves the file pointer to the byte Offset,
// using SeekMode, (which should be either SEEK_SET,
// SEEK_CUR, or SEEK_END).
{
  if (IsOK()) {
     int rv = ::fseek(fp, Offset, SeekMode);
     if (rv != 0)
#ifdef CPP_EXCEPTIONS
       throw CFileSeekError();
#else
       Error->SignalException(EHandler::FileSeekError);
#endif
     LastOperation = SEEK;
  }
  else
#ifdef CPP_EXCEPTIONS
    throw CFileNotReady();
#else
    Error->SignalException(EHandler::FileNotReady);
#endif
}

void VBDFile::Read(void *buf, unsigned Bytes, FAU Address)
// Reads Bytes from Address into buf. The address 
// is always interpreted to be from the beginning of the
// file, unless it's CurrAddress, which means from the
// current position.
{
  if (IsOK()) {
     // Ensure seeks between intervening reads and writes,
     // and optimize for sequential reads
     if (Address == CurrAddress) {
        if (LastOperation == WRITE) Seek(0, SEEK_CUR);
     }
     else Seek(Address, SEEK_SET);
     unsigned bytesmoved = ::fread(buf, 1, Bytes, fp);
     if (bytesmoved != Bytes)
     {
       if(feof(fp))
#ifdef CPP_EXCEPTIONS
	 throw CEOFError();
#else
	 Error->SignalException(EHandler::EOFError);
#endif
       else
#ifdef CPP_EXCEPTIONS
	 throw CFileReadError();
#else
	 Error->SignalException(EHandler::FileReadError);
#endif
     }
     LastOperation = READ;
  }
  else
#ifdef CPP_EXCEPTIONS
    throw CFileNotReady();
#else
    Error->SignalException(EHandler::FileNotReady);
#endif
}

void VBDFile::Write(const void *buf, unsigned Bytes, FAU Address)
// Stores Bytes from buf to Address. The Address is
// always interpreted to be from the beginning of the
// file, unless it's CurrAddress, which means from the
// current position.
{
  if (ReadyForWriting()) {
     // Ensure seeks between intervening reads and writes,
     // and optimize for sequential writes
     if (Address == CurrAddress) {
        if (LastOperation == READ) Seek(0, SEEK_CUR);
     }
     else Seek(Address, SEEK_SET);
     unsigned bytesmoved = ::fwrite(buf, 1, Bytes, fp);
     if (bytesmoved != Bytes)
     {
       if(feof(fp))
#ifdef CPP_EXCEPTIONS
	 throw CEOFError();
#else
	 Error->SignalException(EHandler::EOFError);
#endif
       else
#ifdef CPP_EXCEPTIONS
	 throw CFileWriteError();
#else
	 Error->SignalException(EHandler::FileWriteError);
#endif
     }
     LastOperation = WRITE;
  }
  else
#ifdef CPP_EXCEPTIONS
    throw CFileNotWriteable();
#else
    Error->SignalException(EHandler::FileNotWriteable);
#endif
}

void VBDFile::ReadVBHdr(VBHeader &hdr, FAU Address)
// Reads in VB header, and tests check word to make
// sure the file is still in sync.
{
  Read(&hdr, sizeof(VBHeader), Address);
  if (hdr.CkWord != CheckWord)
#ifdef CPP_EXCEPTIONS
    throw CFileCorrupt();
#else
    Error->SignalException(EHandler::FileCorrupt);
#endif
}

void VBDFile::WriteVBHdr(const VBHeader &hdr, FAU Address)
// Writes the (V)ariable (B)lock header to the file
{
  Write(&hdr, sizeof(VBHeader), Address); // Writes VB header.
}

FAU VBDFile::Reclaim(unsigned Bytes)
// Finds the first block on the free space list that is big
// enough for x Bytes of data. Unused bytes are put back on
// the free list. Reclaim() returns address of the reclaimed
// space, or 0 if there wasn't a deleted VB of the appropriate
// size. Note: the byte size is adjusted by the Alloc()
// function to allocate space for the VB header plus the object.
{
  VBHeader vb, prev_VB, new_VB;
  FAU addr, prev_addr, new_addr;
  unsigned avail_len, unused_len;

  addr = Header.FreeSpace; prev_addr = 0;

  while(addr) { // Until a block of a valid size is found
    ReadVBHdr(vb, addr);
    if (!IsOK()) break;
    avail_len = vb.Length;

    // Unused length must be big enough to hold a two VB headers
    // plus the object
    if (avail_len > Bytes + sizeof(VBHeader)) 
      unused_len = avail_len - Bytes;
    else unused_len = 0;
    
    if (avail_len == Bytes) {
      // Block is an exact fit, so link prev link to Next link
      if (prev_addr == 0) {
	// We're at the head of freespace, so we have a new head
	Header.FreeSpace = vb.NextDeletedVB;
      }
      else {
	// In the middle of free space, so link prev to Next
	prev_VB.NextDeletedVB = vb.NextDeletedVB;
	WriteVBHdr(prev_VB, prev_addr);
      }
      break;
    }
    else if (unused_len > 0) {
      // Block too big, and there's room for a VB header
      // in the unused portion. So splice in this new block.
      new_addr = addr + Bytes;
      new_VB.CkWord = CheckWord;
      new_VB.Status = RemovedVB;
      new_VB.NextDeletedVB = vb.NextDeletedVB;
      new_VB.Length = unused_len;
       WriteVBHdr(new_VB, new_addr);
       if (prev_addr == 0) {
          // We're at the head of freespace, so we have a new head
	 Header.FreeSpace = new_addr;
       }
       else { // In the middle of freespace
	 prev_VB.NextDeletedVB = new_addr;
	 WriteVBHdr(prev_VB, prev_addr);
       }
       break;
    }
    else {
      // Block not big enough, so try Next block
      prev_addr = addr;
      prev_VB = vb;
      addr = vb.NextDeletedVB;
    }
  } // End of looking for big enough block
  
  return IsOK() ? (LWORD)addr : 0;
}

FAU VBDFile::Alloc(unsigned Bytes)
// Allocates a Variable Data Block of x Bytes of data from either
// the free space list, or from the end of the file. The number
// of bytes allocated is adjusted to hold a VBHeader plus the
// object. Only the VBHeader is written to the allocated space.
// Alloc() returns the location of the space allocated for the
// object, or returns a 0 if an error occurred.
{
  FAU Address = 0;
  VBHeader vb;
  
  if (ReadyForWriting()) {

    // Adjust the number of bytes to allocate space of VB header
    Bytes = Bytes + sizeof(VBHeader);
    
    Address = Reclaim(Bytes); // Try using a deleted VB first

    if (IsOK() && Address == 0) { // Extend the file 
      Address = Header.EndOfFile;
      Header.EndOfFile += Bytes;
      // Write to the last byte to avoid possible end of file errors
      SBYTE zero = 0;
      Write(&zero, sizeof(SBYTE), Header.EndOfFile-1);
      Header.HighestVB = Address;  
    }
     vb.Status = NormalVB;  // Mark VB with normal attribute
     vb.NextDeletedVB = 0;  // Always 0 unless VB is marked deleted
     vb.CkWord = CheckWord; // Assign the check word value
     vb.Length = Bytes;     // Total number of bytes for this VB
     
     WriteVBHdr(vb, Address); // Write header for this VB
  }
  else
#ifdef CPP_EXCEPTIONS
    throw CFileNotWriteable();
#else
    Error->SignalException(EHandler::FileNotWriteable);
#endif
  if(Address > Header.HighestVB) Header.HighestVB = Address;  
  return Address + sizeof(VBHeader); // Object's address in the file
}

int VBDFile::Remove(FAU Address)
// Deletes the VB and removes the object by setting all
// the data bytes to zero. The VB is marked removed,
// indicating that the object cannot be undeleted.
{
  // Return false if VB is already deleted or marked bad
  if(!Delete(Address)) return 0;
  
  VBHeader vb;
  FAU VBAddress = Address - sizeof(VBHeader); // Address of VB header
  ReadVBHdr(vb, VBAddress);
  vb.Status = RemovedVB; // Mark VB removed
  unsigned Bytes = vb.Length - sizeof(VBHeader);

  const unsigned len = Bytes;
  char *buf = new char[len+1]; // Create a buffer equal to the object length

  for(unsigned i = 0; i < len; i++) // Zero out the buffer
    buf[i] = 0;
  buf[len] = '\0'; // Ensure the buffer string is null terminated
  
  if (ReadyForWriting()) {
     WriteVBHdr(vb, VBAddress);
     Write(buf, Bytes, Address); // Write all zeros in place of the object
     delete buf;
     return 1; // Return true if successful
  }
  else
#ifdef CPP_EXCEPTIONS
    throw CFileNotWriteable();
#else
    Error->SignalException(EHandler::FileNotWriteable);
#endif
    return 0; // Ensure all paths return a value
}

int VBDFile::Delete(FAU Address)
// Marks the VB at location Address deleted and leaves the
// object unchanged, allowing it to be undeleted. The deleted
// VB is placed on the front of the free space list. 
{
  VBHeader vb;
  
  FAU VBAddress = Address - sizeof(VBHeader); // Address of VB header
  ReadVBHdr(vb, VBAddress);

  // Return false if VB is already deleted
  if(vb.Status != NormalVB) return 0; 
  
  vb.Status = DeletedVB; // Mark VB deleted
  
  if (ReadyForWriting()) {
    vb.NextDeletedVB = Header.FreeSpace; // VB to become head of free list
    WriteVBHdr(vb, VBAddress);
    Header.FreeSpace = VBAddress;
    return 1; // Return true if successful
  }
  else
#ifdef CPP_EXCEPTIONS
    throw CFileNotWriteable();
#else
    Error->SignalException(EHandler::FileNotWriteable);
#endif
    return 0; // Ensure all paths return a value
}

int VBDFile::Exists(const char *FName)
// Returns true if the file exists
{
  FILE *tmp; // Temporary file pointer
  tmp = ::fopen(FName, "rb");

  if(!tmp) return 0; // Return 0 if file does not exist
  
  fclose(tmp);
  return 1;
}

LWORD VBDFile::FileSize(const char *FName)
// Returns the file size. Use after file has been closed
// and re-opened to ensure that all the buffers are flushed
// to disk. 
{
  struct stat buf;
  int result = stat(FName, &buf);
  if(result != 0)
#ifdef CPP_EXCEPTIONS
    throw CFileOpenError();
#else
    Error->SignalException(EHandler::FileOpenError);
#endif
  return buf.st_size;
}

StreamPos VBDFile::FilePosition()
// Returns the current file posistion
{
  StreamPos pos;

  if(!IsOK())
#ifdef CPP_EXCEPTIONS
    throw CFileNotReady();
#else
    Error->SignalException(EHandler::FileNotReady);
#endif

  pos = ::ftell(fp);
  return pos;
}

const char *VBDFile::GetSignature() const
// Return the VBD file signature
{
  size_t len = sizeof(Header.Signature);
  char *s = new char[len +1];
  memcpy(s, Header.Signature, len);
  s[len] = '\0';
  return (const char *)s;
}

unsigned VBDFile::VBTotal()
// Returns the total number of valid VBs in the file
{
  VBHeader vb;
  FAU addr = FindFirstVB(0); // Search the entire file

  unsigned i = 0;
  
  while(addr) { // Until a block of a valid size is found
    if(addr >= Header.EndOfFile) break;
    ReadVBHdr(vb, addr);
    if(!IsOK()) break;
    addr = addr + vb.Length;
    i++;
  }
  return i;
}

unsigned VBDFile::VBDeleted()
// Returns the total number of removed and deleted VBs
{
  VBHeader vb;
  FAU addr = Header.FreeSpace; 
  unsigned i = 0;
  
  while(addr) { // Until a block of a valid size is found
    ReadVBHdr(vb, addr);
    if(!IsOK()) break;
    addr = vb.NextDeletedVB;
    i++;
  }
  return i;
}

unsigned VBDFile::ObjectLength(FAU Address)
// Returns the object length in bytes at the specified address
{
  VBHeader vb;
  FAU VBAddress = Address - sizeof(VBHeader); // Address of VB header
  ReadVBHdr(vb, VBAddress);
  return vb.Length - sizeof(VBHeader);
}

unsigned VBDFile::VBLength(FAU Address)
// Returns the total VB length in bytes at the specified address
{
  VBHeader vb;
  FAU VBAddress = Address - sizeof(VBHeader); // Address of VB header
  ReadVBHdr(vb, VBAddress);
  return vb.Length;
}

int VBDFile::UnDelete(FAU Address)
// Undletes the VB if it has not been removed or reclaimed
{
  VBHeader vb;
  
  FAU VBAddress = Address - sizeof(VBHeader); // Address of VB header
  ReadVBHdr(vb, VBAddress);

  // Return false if VB is not marked deleted 
  if(vb.Status != DeletedVB) return 0; 
  unsigned length = vb.Length;

  // Use Reclaim() to adjust VB header info
  FAU addr = Reclaim(length);

  // Ensure the correct address was reclaimed based on the object length
  if(addr != VBAddress) return 0;

  vb.Status = NormalVB;  // Mark VB with normal attribute
  vb.NextDeletedVB = 0; // Always 0 unless VB is marked deleted
  
  WriteVBHdr(vb, Address - sizeof(VBHeader)); // Write header for this VB

  return 1; // Return true if successful
}

FAU VBDFile::ReAlloc(FAU Address, unsigned Bytes)
// Reallocates a Variable Data Block of x Bytes of data from either
// the free space list, or from the end of the file. ReAlloc()
// returns the location of the space reallocated for the object,
// or returns a 0 if an error occurred.
{
  // Delete the object before allocating space for it
  if(!Delete(Address)) return 0; 

  return Alloc(Bytes);
}

FAU VBDFile::FindFirstVB(FAU Offset)
// Search through the VBD file until a valid VB is found.
// The search starts at the heap start or the offset value.
// Returns 0 if no valid VB is found in the file.
{
  VBHeader vb;
  FAU Address = 0;

  // No VBs have been allocated yet
  if(Header.HeapStart == Header.EndOfFile) return 0;

  if(!Offset)
    Address = Header.HeapStart; // If no Offset, start at heap
  else {
    Address = Address + Offset; // Offset the starting address 

    if(Address >= Header.EndOfFile) // Prevent offsetting past EOF
      return 0; // Invalid address
  }
  
  while(1) {
    if(Address + sizeof(VBHeader) >= Header.EndOfFile || !IsOK()) return 0;
    Read(&vb, sizeof(VBHeader), Address);
    if (vb.CkWord != CheckWord)
      Address++; // Loop through the file byte by byte
    else
      break; // Found valid VB
  }
  return Address;
}

FAU VBDFile::FindFirstObject(FAU Offset)
// Search through the VBD file until a valid VB is found and
// then return the object's address. If the VB is marked deleted
// continue searching until the first normal VB is found. The
// search starts at the heap start or the offset. Returns 0 if
// no valid VB is found in the file.
{
  VBHeader vb;
  FAU Address = FindFirstVB(Offset);
  if(!Address) return 0;

  while(1) { // Loop until a normal VB status is found
    ReadVBHdr(vb, Address);
    if(vb.Status == NormalVB) break;
    Address = FindFirstVB(Address+vb.Length);
    if(!Address) return 0;
  }

  return Address + sizeof(VBHeader);
}

FAU VBDFile::FindNextVB(FAU Offset)
// Search through the VBD file until the next valid VB after the
// first valid VB is found. The search starts at the heap start
// or the offset value. Returns 0 if no valid VB is found.
{
  VBHeader vb;

  // No VBs have been allocated yet
  if(Header.HeapStart == Header.EndOfFile) return 0;

  FAU Address = FindFirstVB(Offset);

  if(!Address) return 0; // No Vaild VB found

  ReadVBHdr(vb, Address);
  FAU NextVB = Address + vb.Length;

  if(NextVB >= Header.EndOfFile) return Address; // This is last the VB

  ReadVBHdr(vb, NextVB); // Ensure VB header is valid

  return NextVB;
}

FAU VBDFile::FindNextObject(FAU Offset)
// Search through the VBD file until the next valid VB after the
// first valid VB is found and then return the object's address.
// If the VB is marked deleted continue searching until the next
// normal VB is found. The search starts at the heap start or the
// offset value. Returns 0 if no valid VB is found.
{
  VBHeader vb;
  FAU Address = FindNextVB(Offset);
  if(!Address) return 0;
  
  while(1) { // Loop until a normal VB status is found
    ReadVBHdr(vb, Address);
    if(vb.Status == NormalVB) break;
    Address = FindNextVB(Address+vb.Length);
    if(!Address) return 0;
  }

  return Address + sizeof(VBHeader);
}

// Debug VBD File manager functions
// ==============================================================
int VBDFile::BlindOpen(const char *FName, AccessMode Mode)
// Opens the FName file without checking the file type.
{
  char *mode_str;

  // First, close the current file if open.
  Close();

  if (Mode == READONLY) {
     mode_str = "rb";
     Status = 0x01; // set good bit, reset read/write bit
  }
  else {
     mode_str = "r+b";
     Status = 0x05; // Set good bit and read/write bit
  }

  fp = ::fopen(FName, mode_str);

  if (fp == 0) {
#ifdef CPP_EXCEPTIONS
    throw CFileOpenError();
#else
    Error->SignalException(EHandler::FileOpenError);
#endif
  }
  else {
    Status |= 2; // Set open bit
    strcpy(FileName, FName);
    LastOperation = WRITE; // So Read() works right the first time
    ReadHdr();
  }

  return IsOpen(); // Returns 1 if file opened successfully, else 0.
}

FAU VBDFile::VBSearch(FAU Offset)
// Search through the VBD file until a valid VB is found.
// The search starts at the beginning  of the file or the
// offset value. Returns 0 if no valid VB is found in the
// file.
{
  VBHeader vb;
  FAU Address = 0;

  FAU StaticEOF = FileSize(VBDFileName());

  if(!Offset)
    Address = sizeof(FileHeader); // If no Offset, start after VBD header
  else {
    Address = Address + Offset; // Offset the starting address 

    if(Address >= StaticEOF) // Prevent offsetting past EOF
      return 0; // Invalid address
  }
  
  while(1) {
    if(Address + sizeof(VBHeader) >= StaticEOF || !IsOK()) return 0;
    Read(&vb, sizeof(VBHeader), Address);
    if (vb.CkWord != CheckWord)
      Address++; // Loop through the file byte by byte
    else
      break; // Found valid VB
  }
  return Address;
}
// ----------------------------------------------------------- //
// ------------------------------- //
// --------- End of File --------- //
// ------------------------------- //
